"""
Exceptions that may be exposed to the users of this library.

"""
import logging
import re
from typing import Optional

import googleapiclient.errors
import httplib2


logger = logging.getLogger(__name__)


class Error(Exception):

    """
    Base class for Google/GCP errors.

    """

    def __repr__(self) -> str:
        return str(self)


class AuthError(Error):

    """
    Google/GCP authentication or authorization error.

    """


class ResourcePermissionDenied(Error):

    """
    Permission denied on a resource.

    For ``HttpError`` exception with "reason":
        "Permission '<permission>' denied for resource '<resource>'."

    """

    def __init__(self, resource: str = None, permission: str = None) -> None:
        """Constructor.

        :param resource: a resource's ID, GRN or some other identifier
        :param permission: a permission's ID, GRN or some other identifier

        """
        self.resource = resource
        self.permission = permission

    def __str__(self) -> str:
        resource_str = self.resource or 'unkwnown'
        permission_str = self.permission or 'unkwnown'

        return "Permission '{permission}' denied for resource '{resource}'.".format(
            resource=resource_str,
            permission=permission_str,
        )


class ResourceNotFound(Error):

    """
    Resource Not Found error.

    For ``HttpError`` exception with "reason":
        "<resource_type> <resource> not found."

    .. warning::
        It might happen that the resource does exist but with the request's
        credentials it is not accessible.

    """

    def __init__(self, resource: str = None) -> None:
        """Constructor.

        :param resource: a resource's ID, GRN or some other identifier

        """
        self.resource = resource

    def __str__(self) -> str:
        resource_str = self.resource or 'unkwnown'

        return "Resource '{resource}' not found.".format(resource=resource_str)


class AlreadyExists(Error):

    """
    Something already exists.

    """

    def __init__(self, what: str) -> None:
        self.what = what

    def __str__(self) -> str:
        return "{what} already exists.".format(what=self.what)


class UnrecognizedApiError(Error):

    """
    Unrecognized Google API ``Error``.

    """

    def __str__(self) -> str:
        return "Unrecognized Google API error."


class UnrecognizedApiHttpError(UnrecognizedApiError):

    """
    Unrecognized Google API ``HttpError``.

    """

    def __init__(self, exc: googleapiclient.errors.HttpError) -> None:
        self.response: httplib2.Response = exc.resp
        self.response_content: bytes = exc.content
        self.request_uri: str = exc.uri
        self.error_reason = "unknown error reason"

        try:
            self.error_reason = exc._get_reason().strip()
        except Exception:
            pass

    def __str__(self) -> str:
        return "Unrecognized Google API HTTP error: {error_reason}.".format(
            error_reason=self.error_reason)


def process_googleapiclient_http_error(
    exc: googleapiclient.errors.HttpError,
) -> Exception:

    new_exc = None
    new_exc = new_exc or _detect_resource_permission_denied(exc)
    new_exc = new_exc or _detect_resource_not_found(exc)
    new_exc = new_exc or _detect_already_exists(exc)

    # TODO: add more cases as they are detected.
    # new_exc = new_exc or _detect_xyz(exc)

    new_exc = new_exc or UnrecognizedApiHttpError(exc)

    return new_exc


def _detect_resource_permission_denied(
    exc: googleapiclient.errors.HttpError,
) -> Optional[Exception]:
    new_exc = None

    try:
        exc_reason = exc._get_reason().strip()
        re_pattern = re.compile(
            r"^Permission '(?P<permission>[a-zA-Z0-9\.]*)' denied for resource '(?P<resource>.*)'\."
        )
        re_match = re_pattern.match(exc_reason)

        if re_match:
            resource = None
            permission = None
            try:
                permission = re_match.group('permission')
            except IndexError:
                pass
            try:
                resource = re_match.group('resource')
            except IndexError:
                pass
            new_exc = ResourcePermissionDenied(resource, permission)
    except Exception:
        pass

    return new_exc


def _detect_resource_not_found(
    exc: googleapiclient.errors.HttpError,
) -> Optional[Exception]:
    new_exc = None

    try:
        exc_reason = exc._get_reason().strip()
        re_pattern = re.compile(r"^(?P<resource_type>[a-zA-Z0-9_]+) (?P<resource>.+) not found\.")
        re_match = re_pattern.match(exc_reason)

        if re_match:
            try:
                resource = re_match.group('resource')
            except IndexError:
                resource = 'some resource'
            new_exc = ResourceNotFound(resource)
    except Exception:
        pass

    return new_exc


def _detect_already_exists(
    exc: googleapiclient.errors.HttpError,
) -> Optional[Exception]:
    new_exc = None

    try:
        exc_reason = exc._get_reason().strip()
        re_pattern = re.compile(r"^(?P<what>.*) already exists\.")
        re_match = re_pattern.match(exc_reason)

        if re_match:
            try:
                what = re_match.group('what')
            except IndexError:
                what = 'something'
            new_exc = AlreadyExists(what)
    except Exception:
        pass

    return new_exc



"""
Definitions and data useful for many services modules.

"""
# warning: do NOT remove any of these definitions, even though they are not used in this module.
from google.auth.credentials import Credentials as GcpCredentials  # noqa: F401
from googleapiclient.discovery import Resource as GcpResource  # noqa: F401



"""
Authentication and authorization utilities.

"""
import logging

import google.auth.compute_engine
import google.auth.credentials
import google.auth.exceptions

from . import exceptions
from .common import GcpCredentials

# Make sure package 'cryptography' is available for 'google.auth', which prefers that lib instead
#   of falling back (silently) to package 'rsa' (pure Python).
#   https://github.com/googleapis/google-auth-library-python/blob/v1.5.1/google/auth/crypt/rsa.py#L19
try:
    import google.auth.crypt._cryptography_rsa
except ImportError as exc:  # pragma: no cover
    msg = "Package 'cryptography' is required for optimum performance of 'google.auth'."
    raise ImportError(msg) from exc


logger = logging.getLogger(__name__)


def get_env_default_credentials() -> GcpCredentials:
    """
    Return the default credentials for the current GCP environment.

    .. warning:: if the env var ``GOOGLE_APPLICATION_CREDENTIALS`` is set, then
        the returned value might correspond to something else.

    """
    try:
        credentials, _ = google.auth.default()
    except google.auth.exceptions.DefaultCredentialsError as exc:
        raise exceptions.AuthError from exc
    return credentials


def get_env_project_id() -> str:
    """
    Return the project ID of the current GCP environment.

    .. warning:: if the env var ``GOOGLE_APPLICATION_CREDENTIALS`` is set, then
        the returned value might correspond to something else.

    """
    try:
        _, project_id = google.auth.default()
    except google.auth.exceptions.DefaultCredentialsError as exc:
        raise exceptions.AuthError from exc
    if not isinstance(project_id, str):
        raise exceptions.Error("Unexpected Google Auth lib response.", project_id)

    return project_id


def get_gce_credentials(service_account_email: str = None) -> GcpCredentials:
    """
    Return credentials provided by Compute Engine service account.

    .. warning:: This function does not attempt to authenticate or verify that
        the ``service_account_email`` does indeed exist. It will return a
        credentials object anyway.

    A Compute Engine instance may have multiple service accounts.

    `Google's Auth Library for Python docs`_ say:

       "Applications running on Compute Engine, Container Engine, or the
       App Engine flexible environment can obtain credentials provided by
       Compute Engine service accounts."

    .. _Google's Auth Library for Python docs:
         https://google-auth.readthedocs.io/en/latest/user-guide.html#compute-engine-container-engine-and-the-app-engine-flexible-environment

    """
    service_account_email = service_account_email or 'default'
    return google.auth.compute_engine.Credentials(service_account_email)


def load_credentials_from_file(filename: str) -> GcpCredentials:
    credentials, _ = google.auth._default._load_credentials_from_file(filename)
    return credentials



"""
A mock version of module :mod:`.gcp_kms`.

The code in this module does not make any external requests.

"""
import base64
import uuid

import cryptography.fernet

from .gcp_kms import GcpCredentials, GcpResource  # noqa: F401
from .gcp_kms import (  # noqa: F401
    compose_crypto_key_grn,
    compose_crypto_key_version_grn,
    compose_key_ring_grn,
    compose_location_grn,
    compose_project_grn,
)
from .gcp_kms import (  # noqa: F401
    KMS_LOCATION_ID_MAX_LENGTH_ESTIMATION,
    KMS_KEY_RING_ID_MAX_LENGTH,
    KMS_KEY_RING_ID_REGEX,
    KMS_CRYPTO_KEY_ID_MAX_LENGTH,
    KMS_CRYPTO_KEY_ID_REGEX,
    KMS_ENCRYPTION_PLAIN_DATA_MAX_SIZE,
)


###############################################################################
# KMS API operations - crypto key
###############################################################################

def create_crypto_key(
    api_client: object,
    key_ring_grn: str,
    crypto_key_id: str = None,
) -> str:
    """
    Create a crypto key (mock) within a key ring.

    Useful for mocking :func:`.gcp_kms.create_crypto_key`.

    :return: crypto key GRN

    """
    # TODO: see TODOs in '.gcp_kms.create_crypto_key'

    crypto_key_id = crypto_key_id or uuid.uuid4().hex
    crypto_key_grn = '{}/cryptoKeys/{}'.format(
        key_ring_grn,
        crypto_key_id,
    )

    return crypto_key_grn


def encrypt(
    api_client: object,
    crypto_key_grn: str,
    plain_data: bytes,
) -> bytes:
    """
    Encrypt binary ``plain_data`` locally, without using GCP KMS.

    Useful for mocking :func:`.gcp_kms.encrypt`.

    """
    if not isinstance(plain_data, bytes):
        raise TypeError("Type of 'plain_data' is not bytes.")
    if len(plain_data) > KMS_ENCRYPTION_PLAIN_DATA_MAX_SIZE:
        raise ValueError("Size of 'plain_data' exceeds max size.")

    fernet_key_input = crypto_key_grn[-32:].encode(encoding='ascii')
    fernet_key = _generate_fernet_key(fernet_key_input)

    f = cryptography.fernet.Fernet(fernet_key)
    encrypted_data: bytes = f.encrypt(plain_data)

    return encrypted_data


def decrypt(
    api_client: object,
    crypto_key_grn: str,
    encrypted_data: bytes,
) -> bytes:
    """
    Decrypt binary ``encrypted_data``locally, without using GCP KMS.

    Useful for mocking :func:`.gcp_kms.decrypt`.

    """
    fernet_key_input = crypto_key_grn[-32:].encode(encoding='ascii')
    fernet_key = _generate_fernet_key(fernet_key_input)

    f = cryptography.fernet.Fernet(fernet_key)
    plain_data: bytes = f.decrypt(encrypted_data)

    return plain_data


###############################################################################
# internal helpers
###############################################################################

def _generate_fernet_key(value: bytes) -> bytes:
    # Based on 'cryptography.fernet.Fernet.generate_key'.
    if not isinstance(value, bytes):
        raise TypeError
    if len(value) != 32:
        raise ValueError
    return base64.urlsafe_b64encode(value)



"""
Fyndata's library of Google Cloud Platform (GCP) utils
======================================================

Among other things, this library makes it easier to interact with some
GCP services or even mock the interaction with them.

Each module that is dedicated to interact with an specific GCP service is named
``gcp_{service_acronym}`` e.g. :mod:`gcp_kms`.

There are other modules for the external users of the library such as
:mod:`exceptions` and :mod:`auth`.


Google Resource Name (GRN)
--------------------------

The terms *resource name* and *resource id* are used extensively in GCP's
documentation and code. Because of the reasons below, for GCP we created an
analogy of AWS' concept of *Amazon Resource Name* (ARN): *Google Resource Name*
(GRN).

- Using the word "name" of X might be misleading since it seems too casual
  and one could easily say "the name of that key ring is ``my-keyring``" but
  that would be wrong: the format of the name of a key ring resource is
  ``projects/[PROJECT_ID]/locations/[LOCATION_ID]/keyRings/[KEY_RING_ID]``.
- It is annoying to have code entities named such as
  ``kms_crypto_key_resource_name`` (unlike ``kms_crypto_key_id``).

"""


__version__ = '0.1.1'



"""
GCP's or this library's-specific common constants.

"""

# GCP project ID
# > A project ID must start with a lowercase letter, and can contain only ASCII letters, digits,
# > and hyphens, and must be between 6 and 30 characters.
# https://cloud.google.com/resource-manager/docs/creating-managing-projects#creating_a_project
PROJECT_ID_MAX_LENGTH = 30
# TODO: PROJECT_ID_REGEX = re.compile(r'^...$')

# GCP region ID
#   https://cloud.google.com/about/locations/
# note: as of 2018-10-22, we have not found anywhere official the value of this restriction.
# note: as of 2018-10-22, the longest region ID is 'northamerica-northeast1' (23 characters).
REGION_ID_MAX_LENGTH_ESTIMATION = 48
# TODO: REGION_ID_MAX_LENGTH



"""
GCP KMS (Key Management Service) helpers.

To learn more about:
- KMS service, see https://cloud.google.com/kms/docs/
- KMS objects hierarchy, see https://cloud.google.com/kms/docs/object-hierarchy
- KMS locations, see https://cloud.google.com/kms/docs/locations

In cryptography the terms for encrypted and not-encrypted values/data are
*ciphertext* and *plaintext*, respectively. However, we choose to put these in
layman's terms: *encrypted data* and *plain data*. We do not want to suggest
that the secret to encrypt is text, which is not (it is a bytes sequence), and
"cipher" is not part of an average software developer's vocabulary.


Usage example, in a GCE environment::

    from fd_gcp.auth import get_gce_credentials

    location_grn = compose_location_grn(
        project_id='[PROJECT_ID]',
        location_id='[LOCATION]',
    )

    credentials = get_gce_credentials()
    kms_api_client = create_api_client(credentials)

    key_ring_grn = create_key_ring(
        kms_api_client, location_grn,
        key_ring_id='test-1')
    crypto_key_grn = create_crypto_key(
        kms_api_client, key_ring_grn,
        crypto_key_id='my-crypto-key-1')

    plain_str = "Jürgen loves Ω! ✔ \n\r\t 123"

    plain_data = plain_str.encode(encoding='utf-8', errors='strict')
    encrypted_data = encrypt(kms_api_client, crypto_key_grn, plain_data)
    decrypted_data = decrypt(kms_api_client, crypto_key_grn, encrypted_data)
    decrypted_str = decrypted_data.decode(encoding='utf-8', errors='strict')

    assert plain_str == decrypted_str


.. seealso:

    https://cloud.google.com/kms/docs/reference/libraries
    https://cloud.google.com/kms/docs/how-tos
    https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.html

"""
import base64
import logging
import re
import uuid
from typing import List

import googleapiclient.discovery

from .common import GcpCredentials, GcpResource
from ._http import execute_google_api_client_request


logger = logging.getLogger(__name__)


###############################################################################
# constants
###############################################################################

# note: as of 2018-10-22, we have not found anywhere official the value of this restriction.
# note: as of 2018-10-22, the longest location id is 'northamerica-northeast1' (23 characters).
# https://cloud.google.com/kms/docs/locations
KMS_LOCATION_ID_MAX_LENGTH_ESTIMATION = 48
# TODO: KMS_LOCATION_ID_MAX_LENGTH
# TODO: KMS_LOCATION_ID_REGEX
# TODO: KMS_LOCATION_GRN_MAX_LENGTH
# TODO: KMS_LOCATION_GRN_REGEX

# > [..] regular expression `[a-zA-Z0-9_-]{1,63}`
# https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.html#create
KMS_KEY_RING_ID_MAX_LENGTH = 64
KMS_KEY_RING_ID_REGEX = re.compile(r'^[a-zA-Z0-9_-]{1,63}$')
# TODO: KMS_KEY_RING_GRN_MAX_LENGTH
# TODO: KMS_KEY_RING_GRN_REGEX

# > [..] regular expression `[a-zA-Z0-9_-]{1,63}`
# https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.cryptoKeys.html#create
KMS_CRYPTO_KEY_ID_MAX_LENGTH = 64
KMS_CRYPTO_KEY_ID_REGEX = re.compile(r'^[a-zA-Z0-9_-]{1,63}$')
# TODO: KMS_CRYPTO_KEY_GRN_MAX_LENGTH
# TODO: KMS_CRYPTO_KEY_GRN_REGEX

# TODO: KMS_CRYPTO_KEY_VERSION_ID_MAX_LENGTH
# TODO: KMS_CRYPTO_KEY_VERSION_ID_REGEX
# TODO: KMS_CRYPTO_KEY_VERSION_GRN_MAX_LENGTH
# TODO: KMS_CRYPTO_KEY_VERSION_GRN_REGEX

# Max size of the data to encrypt with a KMS crypto key.
# > The maximum size depends on the key version's protection_level. For SOFTWARE keys, the plaintext
# > must be no larger than 64KiB. [..]
# https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.cryptoKeys.html#encrypt
KMS_ENCRYPTION_PLAIN_DATA_MAX_SIZE = 64 * 1024  # 64 KiB


###############################################################################
# resource GRN functions
###############################################################################

# The object hierarchy is:
#   'projects/[PROJECT_ID]/locations/[LOCATION]/keyRings/[KEY_RING]/cryptoKeys/[KEY]/cryptoKeyVersions/[VERSION]'

def compose_project_grn(
    project_id: str,
) -> str:
    # TODO: validate params individually or the concat result
    return 'projects/{}'.format(
        project_id,
    )


def compose_location_grn(
    project_id: str,
    location_id: str,
) -> str:
    # TODO: validate params individually or the concat result
    return 'projects/{}/locations/{}'.format(
        project_id,
        location_id,
    )


def compose_key_ring_grn(
    project_id: str,
    location_id: str,
    key_ring_id: str,
) -> str:
    # TODO: validate params individually or the concat result
    return 'projects/{}/locations/{}/keyRings/{}'.format(
        project_id,
        location_id,
        key_ring_id,
    )


def compose_crypto_key_grn(
    project_id: str,
    location_id: str,
    key_ring_id: str,
    crypto_key_id: str,
) -> str:
    # TODO: validate params individually or the concat result
    return 'projects/{}/locations/{}/keyRings/{}/cryptoKeys/{}'.format(
        project_id,
        location_id,
        key_ring_id,
        crypto_key_id,
    )


def compose_crypto_key_version_grn(
    project_id: str,
    location_id: str,
    key_ring_id: str,
    crypto_key_id: str,
    crypto_key_version_id: str,
) -> str:
    # TODO: validate params individually or the concat result
    return 'projects/{}/locations/{}/keyRings/{}/cryptoKeys/{}/cryptoKeyVersions/{}'.format(
        project_id,
        location_id,
        key_ring_id,
        crypto_key_id,
        crypto_key_version_id,
    )


###############################################################################
# KMS API operations
###############################################################################

def create_api_client(credentials: GcpCredentials) -> GcpResource:
    """Create a KMS API client.

    .. warning:: Auth checks do not happen here.

    """
    api_client = googleapiclient.discovery.build(
        serviceName='cloudkms',
        version='v1',
        credentials=credentials,
    )
    return api_client


###############################################################################
# KMS API operations - key ring
###############################################################################

def create_key_ring(
    api_client: GcpResource,
    location_grn: str,
    key_ring_id: str,
) -> str:
    """
    Create a key ring in the given location.

    .. seealso::
        https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.html#create

    Code largely based on
    https://github.com/GoogleCloudPlatform/python-docs-samples/blob/567ef35/kms/api-client/snippets.py

    :return: key ring GRN

    """
    # TODO: validate param 'key_ring_id'

    request = api_client.projects().locations().keyRings().create(
        parent=location_grn,
        body={},
        keyRingId=key_ring_id,
    )
    response = execute_google_api_client_request(request)
    key_ring_grn: str = response['name']

    return key_ring_grn


###############################################################################
# KMS API operations - crypto key
###############################################################################

def create_crypto_key(
    api_client: GcpResource,
    key_ring_grn: str,
    crypto_key_id: str = None,
) -> str:
    """
    Create a crypto key within a key ring.

    .. seealso::
        https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.cryptoKeys.html#create

    Code largely based on
    https://github.com/GoogleCloudPlatform/python-docs-samples/blob/567ef35/kms/api-client/snippets.py

    :return: crypto key GRN

    """
    # TODO: validate param 'crypto_key_id'

    crypto_key_id = crypto_key_id or uuid.uuid4().hex

    request = api_client.projects().locations().keyRings().cryptoKeys().create(
        parent=key_ring_grn,
        body={'purpose': 'ENCRYPT_DECRYPT'},
        cryptoKeyId=crypto_key_id,
    )
    response = execute_google_api_client_request(request)
    crypto_key_grn: str = response['name']

    return crypto_key_grn


def encrypt(
    api_client: GcpResource,
    crypto_key_grn: str,
    plain_data: bytes,
) -> bytes:
    """
    Encrypt binary ``plain_data``.

    .. seealso::
        https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.cryptoKeys.html#encrypt

    Code largely based on
    https://github.com/GoogleCloudPlatform/python-docs-samples/blob/567ef35/kms/api-client/snippets.py

    """
    # TODO: handle encryption/decryption errors

    if not isinstance(plain_data, bytes):
        raise TypeError("Type of 'plain_data' is not bytes.")
    if len(plain_data) > KMS_ENCRYPTION_PLAIN_DATA_MAX_SIZE:
        raise ValueError("Size of 'plain_data' exceeds max size.")

    plain_data_b64_str = base64.b64encode(plain_data).decode('ascii', errors='strict')

    request = api_client.projects().locations().keyRings().cryptoKeys().encrypt(
        name=crypto_key_grn,
        body={'plaintext': plain_data_b64_str},
    )
    response = execute_google_api_client_request(request)
    encrypted_data_b64_str = response['ciphertext']

    encrypted_data = base64.b64decode(encrypted_data_b64_str.encode('ascii', errors='strict'))

    return encrypted_data


def decrypt(
    api_client: GcpResource,
    crypto_key_grn: str,
    encrypted_data: bytes,
) -> bytes:
    """
    Decrypt binary ``encrypted_data``.

    .. seealso::
        https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.cryptoKeys.html#decrypt

    Code largely based on
    https://github.com/GoogleCloudPlatform/python-docs-samples/blob/567ef35/kms/api-client/snippets.py

    """
    # TODO: validate param 'encrypted_data'
    # TODO: handle encryption/decryption errors

    encrypted_data_b64_str = base64.b64encode(encrypted_data).decode('ascii', errors='strict')

    request = api_client.projects().locations().keyRings().cryptoKeys().decrypt(
        name=crypto_key_grn,
        body={'ciphertext': encrypted_data_b64_str},
    )
    response = execute_google_api_client_request(request)
    plain_data_b64_str = response['plaintext']

    plain_data = base64.b64decode(plain_data_b64_str.encode('ascii', errors='strict'))

    return plain_data


###############################################################################
# KMS API operations - crypto key version
###############################################################################

# TODO


###############################################################################
# KMS API operations - IAM policy
###############################################################################

def add_member_to_crypto_key_iam_policy(
    api_client: GcpResource,
    crypto_key_grn: str,
    member: str,
    role: str,
) -> None:
    """
    Add ``member`` with ``role`` to the IAM policy for a crypto key.

    Examples of ``member``:
    - ``user:mike@example.com``
    - ``group:admins@example.com``
    - ``domain:google.com``
    - ``serviceAccount:my-other-app@appspot.gserviceaccount.com``

    Examples of ``role``:
    - ``roles/owner``
    - ``roles/viewer``

    .. seealso::
        https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.cryptoKeys.html#getIamPolicy
        https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.cryptoKeys.html#setIamPolicy

    Code largely based on
    https://github.com/GoogleCloudPlatform/python-docs-samples/blob/567ef35/kms/api-client/snippets.py

    """
    # TODO: validate params 'member', 'role'

    # Get the current IAM policy and add the new member to it.
    policy_request = api_client.projects().locations().keyRings().cryptoKeys().getIamPolicy(
        resource=crypto_key_grn,
    )
    policy_response = execute_google_api_client_request(policy_request)

    bindings: List[dict] = []
    if 'bindings' in policy_response.keys():
        bindings = policy_response['bindings']

    new_binding = {
        'role': role,
        'members': [
            member
        ],
    }
    bindings.append(new_binding)
    policy_response['bindings'] = bindings

    # Set the new IAM Policy.
    request = api_client.projects().locations().keyRings().cryptoKeys().setIamPolicy(
        resource=crypto_key_grn,
        body={'policy': policy_response},
    )
    execute_google_api_client_request(request)


def get_key_ring_iam_policy(
    api_client: GcpResource,
    key_ring_grn: str,
) -> List[dict]:
    """
    Return the IAM policy for a key ring.

    .. seealso::
        https://developers.google.com/resources/api-libraries/documentation/cloudkms/v1/python/latest/cloudkms_v1.projects.locations.keyRings.html#getIamPolicy

    Code largely based on
    https://github.com/GoogleCloudPlatform/python-docs-samples/blob/567ef35/kms/api-client/snippets.py

    :return: the list of bindings of the IAM policy for a key ring

    """
    request = api_client.projects().locations().keyRings().getIamPolicy(
        resource=key_ring_grn,
    )
    response = execute_google_api_client_request(request)

    try:
        bindings = response['bindings']  # type: List[dict]
    except KeyError:
        bindings = []
    return bindings



#!/usr/bin/env python
import os
import re
from typing import Sequence

from setuptools import find_packages, setup


def get_version(*file_paths: Sequence[str]) -> str:
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version('fd_gcp', '__init__.py')


with open('README.md') as readme_file:
    readme = readme_file.read()

# TODO: add reasonable upper-bound per package.
requirements = [
    'cryptography>=2.5',
    'google-api-python-client>=1.7.8',
    'google-auth>=1.6.3',
    'requests>=2.21.0',
]

# extras_requirements = {
# }

setup_requirements = [
]

test_requirements = [
    # note: include here only packages **imported** in test code (e.g. 'requests-mock'), NOT those
    #   like 'coverage' or 'tox'.
]

# note: the "typing information" of this project's packages is not made available to its users
#   automatically; it needs to be packaged and distributed. The way to do so is fairly new and
#   it is specified in PEP 561 - "Distributing and Packaging Type Information".
#   See:
#   - https://www.python.org/dev/peps/pep-0561/#packaging-type-information
#   - https://github.com/python/typing/issues/84
#   - https://github.com/python/mypy/issues/3930
# warning: remember to replicate this in the manifest file for source distribution ('MANIFEST.in').
_package_data = {
    'fd_gcp': [
        # Indicates that the "typing information" of the package should be distributed.
        'py.typed',
    ],
}

setup(
    author='Fyndata (Fynpal SpA)',
    author_email='no-reply@fyndata.com',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
    ],
    description="Fyndata's Python library of Google Cloud Platform (GCP) utils.",
    # extras_require=extras_requirements,
    install_requires=requirements,
    license="MIT",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    name='fyndata-gcp-utils',
    package_data=_package_data,
    packages=find_packages(exclude=['docs', 'tests*']),
    python_requires='>=3.7, <3.8',
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/fyndata/gcp-utils-python',
    version=version,
    zip_safe=False,
)



import logging

import google.auth.exceptions
import googleapiclient.errors
import googleapiclient.http
import httplib2

from . import exceptions


logger = logging.getLogger(__name__)


def execute_google_api_client_request(
    request: googleapiclient.http.HttpRequest,
) -> httplib2.Response:
    try:
        response = request.execute()
    except google.auth.exceptions.GoogleAuthError as exc:
        raise exceptions.AuthError from exc
    except googleapiclient.errors.HttpError as exc:
        new_exc = exceptions.process_googleapiclient_http_error(exc)
        raise new_exc from exc
    except googleapiclient.errors.Error as exc:
        raise exceptions.UnrecognizedApiError from exc

    return response
